# Licensed under a 3-clause BSD style license - see PYFITS.rst

import math
import os
import sys
import time

import numpy as np
import pytest
from numpy.testing import assert_equal

from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning

from .conftest import FitsTestCase


class TestImageFunctions(FitsTestCase):
    def test_constructor_name_arg(self):
        """Like the test of the same name in test_table.py"""

        hdu = fits.ImageHDU()
        assert hdu.name == ""
        assert "EXTNAME" not in hdu.header
        hdu.name = "FOO"
        assert hdu.name == "FOO"
        assert hdu.header["EXTNAME"] == "FOO"

        # Passing name to constructor
        hdu = fits.ImageHDU(name="FOO")
        assert hdu.name == "FOO"
        assert hdu.header["EXTNAME"] == "FOO"

        # And overriding a header with a different extname
        hdr = fits.Header()
        hdr["EXTNAME"] = "EVENTS"
        hdu = fits.ImageHDU(header=hdr, name="FOO")
        assert hdu.name == "FOO"
        assert hdu.header["EXTNAME"] == "FOO"

    def test_constructor_ver_arg(self):
        def assert_ver_is(hdu, reference_ver):
            assert hdu.ver == reference_ver
            assert hdu.header["EXTVER"] == reference_ver

        hdu = fits.ImageHDU()
        assert hdu.ver == 1  # defaults to 1
        assert "EXTVER" not in hdu.header

        hdu.ver = 1
        assert_ver_is(hdu, 1)

        # Passing name to constructor
        hdu = fits.ImageHDU(ver=2)
        assert_ver_is(hdu, 2)

        # And overriding a header with a different extver
        hdr = fits.Header()
        hdr["EXTVER"] = 3
        hdu = fits.ImageHDU(header=hdr, ver=4)
        assert_ver_is(hdu, 4)

        # The header card is not overridden if ver is None or not passed in
        hdr = fits.Header()
        hdr["EXTVER"] = 5
        hdu = fits.ImageHDU(header=hdr, ver=None)
        assert_ver_is(hdu, 5)
        hdu = fits.ImageHDU(header=hdr)
        assert_ver_is(hdu, 5)

    def test_constructor_copies_header(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/153

        Ensure that a header from one HDU is copied when used to initialize new
        HDU.
        """

        ifd = fits.HDUList(fits.PrimaryHDU())
        phdr = ifd[0].header
        phdr["FILENAME"] = "labq01i3q_rawtag.fits"

        primary_hdu = fits.PrimaryHDU(header=phdr)
        ofd = fits.HDUList(primary_hdu)
        ofd[0].header["FILENAME"] = "labq01i3q_flt.fits"

        # Original header should be unchanged
        assert phdr["FILENAME"] == "labq01i3q_rawtag.fits"

    def test_open(self):
        # The function "open" reads a FITS file into an HDUList object.  There
        # are three modes to open: "readonly" (the default), "append", and
        # "update".

        # Open a file read-only (the default mode), the content of the FITS
        # file are read into memory.
        r = fits.open(self.data("test0.fits"))  # readonly

        # data parts are latent instantiation, so if we close the HDUList
        # without touching data, data can not be accessed.
        r.close()

        with pytest.raises(IndexError) as exc_info:
            r[1].data[:2, :2]

        # Check that the exception message is the enhanced version, not the
        # default message from list.__getitem__
        assert str(exc_info.value) == (
            "HDU not found, possibly because the index "
            "is out of range, or because the file was "
            "closed before all HDUs were read"
        )

    def test_open_2(self):
        r = fits.open(self.data("test0.fits"))

        info = [(0, "PRIMARY", 1, "PrimaryHDU", 138, (), "", "")] + [
            (x, "SCI", x, "ImageHDU", 61, (40, 40), "int16", "") for x in range(1, 5)
        ]

        try:
            assert r.info(output=False) == info
        finally:
            r.close()

    def test_open_3(self):
        # Test that HDUs cannot be accessed after the file was closed
        r = fits.open(self.data("test0.fits"))
        r.close()
        with pytest.raises(IndexError) as exc_info:
            r[1]

        # Check that the exception message is the enhanced version, not the
        # default message from list.__getitem__
        assert str(exc_info.value) == (
            "HDU not found, possibly because the index "
            "is out of range, or because the file was "
            "closed before all HDUs were read"
        )

        # Test that HDUs can be accessed with lazy_load_hdus=False
        r = fits.open(self.data("test0.fits"), lazy_load_hdus=False)
        r.close()
        assert isinstance(r[1], fits.ImageHDU)
        assert len(r) == 5

        with pytest.raises(IndexError) as exc_info:
            r[6]
        assert str(exc_info.value) == "list index out of range"

        # And the same with the global config item
        assert fits.conf.lazy_load_hdus  # True by default
        fits.conf.lazy_load_hdus = False
        try:
            r = fits.open(self.data("test0.fits"))
            r.close()
            assert isinstance(r[1], fits.ImageHDU)
            assert len(r) == 5
        finally:
            fits.conf.lazy_load_hdus = True

    def test_fortran_array(self):
        # Test that files are being correctly written+read for "C" and "F" order arrays
        a = np.arange(21).reshape(3, 7)
        b = np.asfortranarray(a)

        afits = self.temp("a_str.fits")
        bfits = self.temp("b_str.fits")
        # writing to str specified files
        fits.PrimaryHDU(data=a).writeto(afits)
        fits.PrimaryHDU(data=b).writeto(bfits)
        np.testing.assert_array_equal(fits.getdata(afits), a)
        np.testing.assert_array_equal(fits.getdata(bfits), a)

        # writing to fileobjs
        aafits = self.temp("a_fileobj.fits")
        bbfits = self.temp("b_fileobj.fits")
        with open(aafits, mode="wb") as fd:
            fits.PrimaryHDU(data=a).writeto(fd)
        with open(bbfits, mode="wb") as fd:
            fits.PrimaryHDU(data=b).writeto(fd)
        np.testing.assert_array_equal(fits.getdata(aafits), a)
        np.testing.assert_array_equal(fits.getdata(bbfits), a)

    def test_fortran_array_non_contiguous(self):
        # Test that files are being correctly written+read for 'C' and 'F' order arrays
        a = np.arange(105).reshape(3, 5, 7)
        b = np.asfortranarray(a)

        # writing to str specified files
        afits = self.temp("a_str_slice.fits")
        bfits = self.temp("b_str_slice.fits")
        fits.PrimaryHDU(data=a[::2, ::2]).writeto(afits)
        fits.PrimaryHDU(data=b[::2, ::2]).writeto(bfits)
        np.testing.assert_array_equal(fits.getdata(afits), a[::2, ::2])
        np.testing.assert_array_equal(fits.getdata(bfits), a[::2, ::2])

        # writing to fileobjs
        aafits = self.temp("a_fileobj_slice.fits")
        bbfits = self.temp("b_fileobj_slice.fits")
        with open(aafits, mode="wb") as fd:
            fits.PrimaryHDU(data=a[::2, ::2]).writeto(fd)
        with open(bbfits, mode="wb") as fd:
            fits.PrimaryHDU(data=b[::2, ::2]).writeto(fd)
        np.testing.assert_array_equal(fits.getdata(aafits), a[::2, ::2])
        np.testing.assert_array_equal(fits.getdata(bbfits), a[::2, ::2])

    def test_primary_with_extname(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/151

        Tests that the EXTNAME keyword works with Primary HDUs as well, and
        interacts properly with the .name attribute.  For convenience
        hdulist['PRIMARY'] will still refer to the first HDU even if it has an
        EXTNAME not equal to 'PRIMARY'.
        """

        prihdr = fits.Header([("EXTNAME", "XPRIMARY"), ("EXTVER", 1)])
        hdul = fits.HDUList([fits.PrimaryHDU(header=prihdr)])
        assert "EXTNAME" in hdul[0].header
        assert hdul[0].name == "XPRIMARY"
        assert hdul[0].name == hdul[0].header["EXTNAME"]

        info = [(0, "XPRIMARY", 1, "PrimaryHDU", 5, (), "", "")]
        assert hdul.info(output=False) == info

        assert hdul["PRIMARY"] is hdul["XPRIMARY"]
        assert hdul["PRIMARY"] is hdul[("XPRIMARY", 1)]

        hdul[0].name = "XPRIMARY2"
        assert hdul[0].header["EXTNAME"] == "XPRIMARY2"

        hdul.writeto(self.temp("test.fits"))
        with fits.open(self.temp("test.fits")) as hdul:
            assert hdul[0].name == "XPRIMARY2"

    @pytest.mark.filterwarnings(
        "ignore:Memory map object was closed but appears to still be referenced:UserWarning"
    )
    def test_io_manipulation(self):
        # Get a keyword value.  An extension can be referred by name or by
        # number.  Both extension and keyword names are case insensitive.
        with fits.open(self.data("test0.fits")) as r:
            assert r["primary"].header["naxis"] == 0
            assert r[0].header["naxis"] == 0

            # If there are more than one extension with the same EXTNAME value,
            # the EXTVER can be used (as the second argument) to distinguish
            # the extension.
            assert r["sci", 1].header["detector"] == 1

            # append (using "update()") a new card
            r[0].header["xxx"] = 1.234e56

            assert (
                "\n".join(str(x) for x in r[0].header.cards[-3:])
                == "EXPFLAG = 'NORMAL            ' / Exposure interruption indicator                \n"
                "FILENAME= 'vtest3.fits'        / File name                                      \n"
                "XXX     =            1.234E+56                                                  "
            )

            # rename a keyword
            r[0].header.rename_keyword("filename", "fname")
            pytest.raises(ValueError, r[0].header.rename_keyword, "fname", "history")

            pytest.raises(ValueError, r[0].header.rename_keyword, "fname", "simple")
            r[0].header.rename_keyword("fname", "filename")

            # get a subsection of data
            assert np.array_equal(
                r[2].data[:3, :3],
                np.array(
                    [[349, 349, 348], [349, 349, 347], [347, 350, 349]], dtype=np.int16
                ),
            )

            # We can create a new FITS file by opening a new file with "append"
            # mode.
            with fits.open(self.temp("test_new.fits"), mode="append") as n:
                # Append the primary header and the 2nd extension to the new
                # file.
                n.append(r[0])
                n.append(r[2])

                # The flush method will write the current HDUList object back
                # to the newly created file on disk.  The HDUList is still open
                # and can be further operated.
                n.flush()
                assert n[1].data[1, 1] == 349

                # modify a data point
                n[1].data[1, 1] = 99

                # When the file is closed, the most recent additions of
                # extension(s) since last flush() will be appended, but any HDU
                # already existed at the last flush will not be modified
            del n

            # If an existing file is opened with "append" mode, like the
            # readonly mode, the HDU's will be read into the HDUList which can
            # be modified in memory but can not be written back to the original
            # file.  A file opened with append mode can only add new HDU's.
            os.rename(self.temp("test_new.fits"), self.temp("test_append.fits"))

            with fits.open(self.temp("test_append.fits"), mode="append") as a:
                # The above change did not take effect since this was made
                # after the flush().
                assert a[1].data[1, 1] == 349
                a.append(r[1])
            del a

            # When changes are made to an HDUList which was opened with
            # "update" mode, they will be written back to the original file
            # when a flush/close is called.
            os.rename(self.temp("test_append.fits"), self.temp("test_update.fits"))

            with fits.open(self.temp("test_update.fits"), mode="update") as u:
                # When the changes do not alter the size structures of the
                # original (or since last flush) HDUList, the changes are
                # written back "in place".
                assert u[0].header["rootname"] == "U2EQ0201T"
                u[0].header["rootname"] = "abc"
                assert u[1].data[1, 1] == 349
                u[1].data[1, 1] = 99
                u.flush()

                # If the changes affect the size structure, e.g. adding or
                # deleting HDU(s), header was expanded or reduced beyond
                # existing number of blocks (2880 bytes in each block), or
                # change the data size, the HDUList is written to a temporary
                # file, the original file is deleted, and the temporary file is
                # renamed to the original file name and reopened in the update
                # mode.  To a user, these two kinds of updating writeback seem
                # to be the same, unless the optional argument in flush or
                # close is set to 1.
                del u[2]
                u.flush()

                # The write method in HDUList class writes the current HDUList,
                # with all changes made up to now, to a new file.  This method
                # works the same disregard the mode the HDUList was opened
                # with.
                u.append(r[3])
                u.writeto(self.temp("test_new.fits"))
            del u

        # Another useful new HDUList method is readall.  It will "touch" the
        # data parts in all HDUs, so even if the HDUList is closed, we can
        # still operate on the data.
        with fits.open(self.data("test0.fits")) as r:
            r.readall()
            assert r[1].data[1, 1] == 315

        # create an HDU with data only
        data = np.ones((3, 5), dtype=np.float32)
        hdu = fits.ImageHDU(data=data, name="SCI")
        assert np.array_equal(
            hdu.data,
            np.array(
                [
                    [1.0, 1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                dtype=np.float32,
            ),
        )

        # create an HDU with header and data
        # notice that the header has the right NAXIS's since it is constructed
        # with ImageHDU
        hdu2 = fits.ImageHDU(header=r[1].header, data=np.array([1, 2], dtype="int32"))

        assert (
            "\n".join(str(x) for x in hdu2.header.cards[1:5])
            == "BITPIX  =                   32 / array data type                                \n"
            "NAXIS   =                    1 / number of array dimensions                     \n"
            "NAXIS1  =                    2                                                  \n"
            "PCOUNT  =                    0 / number of parameters                           "
        )

    def test_memory_mapping(self):
        # memory mapping
        f1 = fits.open(self.data("test0.fits"), memmap=1)
        f1.close()

    def test_verification_on_output(self):
        # verification on output
        # make a defect HDUList first
        x = fits.ImageHDU()
        hdu = fits.HDUList(x)  # HDUList can take a list or one single HDU

        with pytest.warns(fits.verify.VerifyWarning) as w:
            hdu.verify()
        assert len(w) == 3
        assert "HDUList's 0th element is not a primary HDU" in str(w[1].message)

        with pytest.warns(fits.verify.VerifyWarning) as w:
            hdu.writeto(self.temp("test_new2.fits"), "fix")
        assert len(w) == 3
        assert "Fixed by inserting one as 0th HDU" in str(w[1].message)

    def test_section(self):
        # section testing
        fs = fits.open(self.data("arange.fits"))
        assert np.array_equal(fs[0].section[3, 2, 5], 357)
        assert np.array_equal(
            fs[0].section[3, 2, :],
            np.array([352, 353, 354, 355, 356, 357, 358, 359, 360, 361, 362]),
        )
        assert np.array_equal(
            fs[0].section[3, 2, 4:], np.array([356, 357, 358, 359, 360, 361, 362])
        )
        assert np.array_equal(
            fs[0].section[3, 2, :8], np.array([352, 353, 354, 355, 356, 357, 358, 359])
        )
        assert np.array_equal(
            fs[0].section[3, 2, -8:8], np.array([355, 356, 357, 358, 359])
        )
        assert np.array_equal(
            fs[0].section[3, 2:5, :],
            np.array(
                [
                    [352, 353, 354, 355, 356, 357, 358, 359, 360, 361, 362],
                    [363, 364, 365, 366, 367, 368, 369, 370, 371, 372, 373],
                    [374, 375, 376, 377, 378, 379, 380, 381, 382, 383, 384],
                ]
            ),
        )

        assert np.array_equal(
            fs[0].section[3, :, :][:3, :3],
            np.array([[330, 331, 332], [341, 342, 343], [352, 353, 354]]),
        )

        dat = fs[0].data
        assert np.array_equal(fs[0].section[3, 2:5, :8], dat[3, 2:5, :8])
        assert np.array_equal(fs[0].section[3, 2:5, 3], dat[3, 2:5, 3])

        assert np.array_equal(
            fs[0].section[3:6, :, :][:3, :3, :3],
            np.array(
                [
                    [[330, 331, 332], [341, 342, 343], [352, 353, 354]],
                    [[440, 441, 442], [451, 452, 453], [462, 463, 464]],
                    [[550, 551, 552], [561, 562, 563], [572, 573, 574]],
                ]
            ),
        )

        assert np.array_equal(
            fs[0].section[:, :, :][:3, :2, :2],
            np.array(
                [[[0, 1], [11, 12]], [[110, 111], [121, 122]], [[220, 221], [231, 232]]]
            ),
        )

        assert np.array_equal(fs[0].section[:, 2, :], dat[:, 2, :])
        assert np.array_equal(fs[0].section[:, 2:5, :], dat[:, 2:5, :])
        assert np.array_equal(fs[0].section[3:6, 3, :], dat[3:6, 3, :])
        assert np.array_equal(fs[0].section[3:6, 3:7, :], dat[3:6, 3:7, :])

        assert np.array_equal(fs[0].section[:, ::2], dat[:, ::2])
        assert np.array_equal(fs[0].section[:, [1, 2, 4], 3], dat[:, [1, 2, 4], 3])
        bool_index = np.array(
            [True, False, True, True, False, False, True, True, False, True]
        )
        assert np.array_equal(fs[0].section[:, bool_index, :], dat[:, bool_index, :])

        assert np.array_equal(fs[0].section[3:6, 3, :, ...], dat[3:6, 3, :, ...])
        assert np.array_equal(fs[0].section[..., ::2], dat[..., ::2])
        assert np.array_equal(fs[0].section[..., [1, 2, 4], 3], dat[..., [1, 2, 4], 3])

        # Can we use negative indices?
        assert np.array_equal(fs[0].section[-1], dat[-1])
        assert np.array_equal(fs[0].section[-9:-7], dat[-9:-7])
        assert np.array_equal(fs[0].section[-4, -6:-3, -1], dat[-4, -6:-3, -1])
        fs.close()

    def test_section_data_single(self):
        a = np.array([1])
        hdu = fits.PrimaryHDU(a)
        hdu.writeto(self.temp("test_new.fits"))

        hdul = fits.open(self.temp("test_new.fits"))
        sec = hdul[0].section
        dat = hdul[0].data
        assert np.array_equal(sec[0], dat[0])
        assert np.array_equal(sec[...], dat[...])
        assert np.array_equal(sec[..., 0], dat[..., 0])
        assert np.array_equal(sec[0, ...], dat[0, ...])
        hdul.close()

    def test_section_data_square(self):
        a = np.arange(4).reshape(2, 2)
        hdu = fits.PrimaryHDU(a)
        hdu.writeto(self.temp("test_new.fits"))

        hdul = fits.open(self.temp("test_new.fits"))
        d = hdul[0]
        dat = hdul[0].data
        assert (d.section[:, :] == dat[:, :]).all()
        assert (d.section[0, :] == dat[0, :]).all()
        assert (d.section[1, :] == dat[1, :]).all()
        assert (d.section[:, 0] == dat[:, 0]).all()
        assert (d.section[:, 1] == dat[:, 1]).all()
        assert (d.section[0, 0] == dat[0, 0]).all()
        assert (d.section[0, 1] == dat[0, 1]).all()
        assert (d.section[1, 0] == dat[1, 0]).all()
        assert (d.section[1, 1] == dat[1, 1]).all()
        assert (d.section[0:1, 0:1] == dat[0:1, 0:1]).all()
        assert (d.section[0:2, 0:1] == dat[0:2, 0:1]).all()
        assert (d.section[0:1, 0:2] == dat[0:1, 0:2]).all()
        assert (d.section[0:2, 0:2] == dat[0:2, 0:2]).all()
        hdul.close()

    def test_section_data_cube(self):
        a = np.arange(18).reshape(2, 3, 3)
        hdu = fits.PrimaryHDU(a)
        hdu.writeto(self.temp("test_new.fits"))

        hdul = fits.open(self.temp("test_new.fits"))
        d = hdul[0]
        dat = hdul[0].data

        assert (d.section[:] == dat[:]).all()
        assert (d.section[:, :] == dat[:, :]).all()

        # Test that various combinations of indexing on the section are equal to
        # indexing the data.
        # Testing all combinations of scalar-index and [:] for each dimension.
        for idx1 in [slice(None), 0, 1]:
            for idx2 in [slice(None), 0, 1, 2]:
                for idx3 in [slice(None), 0, 1, 2]:
                    nd_idx = (idx1, idx2, idx3)
                    assert (d.section[nd_idx] == dat[nd_idx]).all()

        # Test all ways to slice the last dimension but keeping the first two.
        for idx3 in [
            slice(0, 1),
            slice(0, 2),
            slice(0, 3),
            slice(1, 2),
            slice(1, 3),
            slice(2, 3),
        ]:
            nd_idx = (slice(None), slice(None), idx3)
            assert (d.section[nd_idx] == dat[nd_idx]).all()

        # Test various combinations (not exhaustive) to slice all dimensions.
        for idx1 in [slice(0, 1), slice(1, 2)]:
            for idx2 in [
                slice(0, 1),
                slice(0, 2),
                slice(0, 3),
                slice(1, 2),
                slice(1, 3),
            ]:
                for idx3 in [
                    slice(0, 1),
                    slice(0, 2),
                    slice(0, 3),
                    slice(1, 2),
                    slice(1, 3),
                    slice(2, 3),
                ]:
                    nd_idx = (idx1, idx2, idx3)
                    assert (d.section[nd_idx] == dat[nd_idx]).all()

        hdul.close()

    def test_section_data_four(self):
        a = np.arange(256).reshape(4, 4, 4, 4)
        hdu = fits.PrimaryHDU(a)
        hdu.writeto(self.temp("test_new.fits"))

        hdul = fits.open(self.temp("test_new.fits"))
        d = hdul[0]
        dat = hdul[0].data
        assert (d.section[:, :, :, :] == dat[:, :, :, :]).all()
        assert (d.section[:, :, :] == dat[:, :, :]).all()
        assert (d.section[:, :] == dat[:, :]).all()
        assert (d.section[:] == dat[:]).all()
        assert (d.section[0, :, :, :] == dat[0, :, :, :]).all()
        assert (d.section[0, :, 0, :] == dat[0, :, 0, :]).all()
        assert (d.section[:, :, 0, :] == dat[:, :, 0, :]).all()
        assert (d.section[:, 1, 0, :] == dat[:, 1, 0, :]).all()
        assert (d.section[:, :, :, 1] == dat[:, :, :, 1]).all()
        hdul.close()

    def test_section_data_scaled(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/143

        This is like test_section_data_square but uses a file containing scaled
        image data, to test that sections can work correctly with scaled data.
        """

        hdul = fits.open(self.data("scale.fits"))
        d = hdul[0]
        dat = hdul[0].data
        assert (d.section[:, :] == dat[:, :]).all()
        assert (d.section[0, :] == dat[0, :]).all()
        assert (d.section[1, :] == dat[1, :]).all()
        assert (d.section[:, 0] == dat[:, 0]).all()
        assert (d.section[:, 1] == dat[:, 1]).all()
        assert (d.section[0, 0] == dat[0, 0]).all()
        assert (d.section[0, 1] == dat[0, 1]).all()
        assert (d.section[1, 0] == dat[1, 0]).all()
        assert (d.section[1, 1] == dat[1, 1]).all()
        assert (d.section[0:1, 0:1] == dat[0:1, 0:1]).all()
        assert (d.section[0:2, 0:1] == dat[0:2, 0:1]).all()
        assert (d.section[0:1, 0:2] == dat[0:1, 0:2]).all()
        assert (d.section[0:2, 0:2] == dat[0:2, 0:2]).all()
        hdul.close()

        # Test without having accessed the full data first
        hdul = fits.open(self.data("scale.fits"))
        d = hdul[0]
        assert (d.section[:, :] == dat[:, :]).all()
        assert (d.section[0, :] == dat[0, :]).all()
        assert (d.section[1, :] == dat[1, :]).all()
        assert (d.section[:, 0] == dat[:, 0]).all()
        assert (d.section[:, 1] == dat[:, 1]).all()
        assert (d.section[0, 0] == dat[0, 0]).all()
        assert (d.section[0, 1] == dat[0, 1]).all()
        assert (d.section[1, 0] == dat[1, 0]).all()
        assert (d.section[1, 1] == dat[1, 1]).all()
        assert (d.section[0:1, 0:1] == dat[0:1, 0:1]).all()
        assert (d.section[0:2, 0:1] == dat[0:2, 0:1]).all()
        assert (d.section[0:1, 0:2] == dat[0:1, 0:2]).all()
        assert (d.section[0:2, 0:2] == dat[0:2, 0:2]).all()
        assert not d._data_loaded
        hdul.close()

    def test_do_not_scale_image_data(self):
        with fits.open(self.data("scale.fits"), do_not_scale_image_data=True) as hdul:
            assert hdul[0].data.dtype == np.dtype(">i2")

        with fits.open(self.data("scale.fits")) as hdul:
            assert hdul[0].data.dtype == np.dtype("float32")

    def test_append_uint_data(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/56
        (BZERO and BSCALE added in the wrong location when appending scaled
        data)
        """

        fits.writeto(self.temp("test_new.fits"), data=np.array([], dtype="uint8"))
        d = np.zeros([100, 100]).astype("uint16")
        fits.append(self.temp("test_new.fits"), data=d)

        with fits.open(self.temp("test_new.fits"), uint=True) as f:
            assert f[1].data.dtype == "uint16"

    def test_scale_with_explicit_bzero_bscale(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/6399
        """
        hdu2 = fits.ImageHDU(np.random.rand(100, 100))
        # The line below raised an exception in astropy 2.0, so if it does not
        # raise an error here, that is progress.
        hdu2.scale(type="uint8", bscale=1, bzero=0)

    def test_uint_header_consistency(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2305

        This ensures that an HDU containing unsigned integer data always has
        the appropriate BZERO value in its header.
        """

        for int_size in (16, 32, 64):
            # Just make an array of some unsigned ints that wouldn't fit in a
            # signed int array of the same bit width
            max_uint = (2**int_size) - 1
            if int_size == 64:
                max_uint = np.uint64(int_size)

            dtype = f"uint{int_size}"
            arr = np.empty(100, dtype=dtype)
            arr.fill(max_uint)
            arr -= np.arange(100, dtype=dtype)

            uint_hdu = fits.PrimaryHDU(data=arr)
            assert np.all(uint_hdu.data == arr)
            assert uint_hdu.data.dtype.name == f"uint{int_size}"
            assert "BZERO" in uint_hdu.header
            assert uint_hdu.header["BZERO"] == (2 ** (int_size - 1))

            filename = f"uint{int_size}.fits"
            uint_hdu.writeto(self.temp(filename))

            with fits.open(self.temp(filename), uint=True) as hdul:
                new_uint_hdu = hdul[0]
                assert np.all(new_uint_hdu.data == arr)
                assert new_uint_hdu.data.dtype.name == f"uint{int_size}"
                assert "BZERO" in new_uint_hdu.header
                assert new_uint_hdu.header["BZERO"] == (2 ** (int_size - 1))

    @pytest.mark.parametrize(("from_file"), (False, True))
    @pytest.mark.parametrize(("do_not_scale"), (False,))
    def test_uint_header_keywords_removed_after_bitpix_change(
        self, from_file, do_not_scale
    ):
        """
        Regression test for https://github.com/astropy/astropy/issues/4974

        BZERO/BSCALE should be removed if data is converted to a floating
        point type.

        Currently excluding the case where do_not_scale_image_data=True
        because it is not clear what the expectation should be.
        """

        arr = np.zeros(100, dtype="uint16")

        if from_file:
            # To generate the proper input file we always want to scale the
            # data before writing it...otherwise when we open it will be
            # regular (signed) int data.
            tmp_uint = fits.PrimaryHDU(arr)
            filename = "unsigned_int.fits"
            tmp_uint.writeto(self.temp(filename))
            with fits.open(
                self.temp(filename), do_not_scale_image_data=do_not_scale
            ) as f:
                uint_hdu = f[0]
                # Force a read before we close.
                _ = uint_hdu.data
        else:
            uint_hdu = fits.PrimaryHDU(arr, do_not_scale_image_data=do_not_scale)

        # Make sure appropriate keywords are in the header. See
        # https://github.com/astropy/astropy/pull/3916#issuecomment-122414532
        # for discussion.
        assert "BSCALE" in uint_hdu.header
        assert "BZERO" in uint_hdu.header
        assert uint_hdu.header["BSCALE"] == 1
        assert uint_hdu.header["BZERO"] == 32768

        # Convert data to floating point...
        uint_hdu.data = uint_hdu.data * 1.0

        # ...bitpix should be negative.
        assert uint_hdu.header["BITPIX"] < 0

        # BSCALE and BZERO should NOT be in header any more.
        assert "BSCALE" not in uint_hdu.header
        assert "BZERO" not in uint_hdu.header

        # This is the main test...the data values should round trip
        # as zero.
        filename = "test_uint_to_float.fits"
        uint_hdu.writeto(self.temp(filename))
        with fits.open(self.temp(filename)) as hdul:
            assert (hdul[0].data == 0).all()

    def test_blanks(self):
        """Test image data with blank spots in it (which should show up as
        NaNs in the data array.
        """

        arr = np.zeros((10, 10), dtype=np.int32)
        # One row will be blanks
        arr[1] = 999
        hdu = fits.ImageHDU(data=arr)
        hdu.header["BLANK"] = 999
        hdu.writeto(self.temp("test_new.fits"))

        with fits.open(self.temp("test_new.fits")) as hdul:
            assert np.isnan(hdul[1].data[1]).all()

    def test_invalid_blanks(self):
        """
        Test that invalid use of the BLANK keyword leads to an appropriate
        warning, and that the BLANK keyword is ignored when returning the
        HDU data.

        Regression test for https://github.com/astropy/astropy/issues/3865
        """

        arr = np.arange(5, dtype=np.float64)
        hdu = fits.PrimaryHDU(data=arr)
        hdu.header["BLANK"] = 2

        with pytest.warns(
            AstropyUserWarning, match="Invalid 'BLANK' keyword in header"
        ) as w:
            hdu.writeto(self.temp("test_new.fits"))
        # Allow the HDU to be written, but there should be a warning
        # when writing a header with BLANK when then data is not
        # int
        assert len(w) == 1

        # Should also get a warning when opening the file, and the BLANK
        # value should not be applied
        with pytest.warns(
            AstropyUserWarning, match="Invalid 'BLANK' keyword in header"
        ) as w:
            with fits.open(self.temp("test_new.fits")) as h:
                assert np.all(arr == h[0].data)
        assert len(w) == 1

    @pytest.mark.filterwarnings("ignore:Invalid 'BLANK' keyword in header")
    def test_scale_back_with_blanks(self):
        """
        Test that when auto-rescaling integer data with "blank" values (where
        the blanks are replaced by NaN in the float data), that the "BLANK"
        keyword is removed from the header.

        Further, test that when using the ``scale_back=True`` option the blank
        values are restored properly.

        Regression test for https://github.com/astropy/astropy/issues/3865
        """

        # Make the sample file
        arr = np.arange(5, dtype=np.int32)
        hdu = fits.PrimaryHDU(data=arr)
        hdu.scale("int16", bscale=1.23)

        # Creating data that uses BLANK is currently kludgy--a separate issue
        # TODO: Rewrite this test when scaling with blank support is better
        # supported

        # Let's just add a value to the data that should be converted to NaN
        # when it is read back in:
        filename = self.temp("test.fits")
        hdu.data[0] = 9999
        hdu.header["BLANK"] = 9999
        hdu.writeto(filename)

        with fits.open(filename) as hdul:
            data = hdul[0].data
            assert np.isnan(data[0])
            with pytest.warns(
                fits.verify.VerifyWarning, match=r"Invalid 'BLANK' keyword in header"
            ):
                hdul.writeto(self.temp("test2.fits"))

        # Now reopen the newly written file.  It should not have a 'BLANK'
        # keyword
        with fits.open(self.temp("test2.fits")) as hdul2:
            assert "BLANK" not in hdul2[0].header
            data = hdul2[0].data
            assert np.isnan(data[0])

        # Finally, test that scale_back keeps the BLANKs correctly
        with fits.open(filename, scale_back=True, mode="update") as hdul3:
            data = hdul3[0].data
            # This emits warning that pytest cannot catch properly, so we
            # catch it with pytest.mark.filterwarnings above.
            assert np.isnan(data[0])

        with fits.open(filename, do_not_scale_image_data=True) as hdul4:
            assert hdul4[0].header["BLANK"] == 9999
            assert hdul4[0].header["BSCALE"] == 1.23
            assert hdul4[0].data[0] == 9999

    def test_bzero_with_floats(self):
        """Test use of the BZERO keyword in an image HDU containing float
        data.
        """

        arr = np.zeros((10, 10)) - 1
        hdu = fits.ImageHDU(data=arr)
        hdu.header["BZERO"] = 1.0
        hdu.writeto(self.temp("test_new.fits"))

        with fits.open(self.temp("test_new.fits")) as hdul:
            arr += 1
            assert (hdul[1].data == arr).all()

    def test_rewriting_large_scaled_image(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/84 and
        https://aeon.stsci.edu/ssb/trac/pyfits/ticket/101
        """

        hdul = fits.open(self.data("fixed-1890.fits"))
        orig_data = hdul[0].data
        hdul.writeto(self.temp("test_new.fits"), overwrite=True)
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[0].data == orig_data).all()
        hdul.close()

        # Just as before, but this time don't touch hdul[0].data before writing
        # back out--this is the case that failed in
        # https://aeon.stsci.edu/ssb/trac/pyfits/ticket/84
        hdul = fits.open(self.data("fixed-1890.fits"))
        hdul.writeto(self.temp("test_new.fits"), overwrite=True)
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[0].data == orig_data).all()
        hdul.close()

        # Test opening/closing/reopening a scaled file in update mode
        hdul = fits.open(self.data("fixed-1890.fits"), do_not_scale_image_data=True)
        hdul.writeto(
            self.temp("test_new.fits"), overwrite=True, output_verify="silentfix"
        )
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        orig_data = hdul[0].data
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"), mode="update")
        hdul.close()
        hdul = fits.open(self.temp("test_new.fits"))
        assert (hdul[0].data == orig_data).all()
        hdul.close()

    def test_image_update_header(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/105

        Replacing the original header to an image HDU and saving should update
        the NAXISn keywords appropriately and save the image data correctly.
        """

        # Copy the original file before saving to it
        self.copy_file("test0.fits")
        with fits.open(self.temp("test0.fits"), mode="update") as hdul:
            orig_data = hdul[1].data.copy()
            hdr_copy = hdul[1].header.copy()
            del hdr_copy["NAXIS*"]
            hdul[1].header = hdr_copy

        with fits.open(self.temp("test0.fits")) as hdul:
            assert (orig_data == hdul[1].data).all()

    if sys.version_info < (3, 11):
        # The test below raised a `ResourceWarning: unclosed transport` exception
        # due to a bug in Python < 3.11 (cf. cpython#90476)
        mark = pytest.mark.filterwarnings("ignore:unclosed transport <asyncio.sslproto")
    else:
        mark = lambda f: f

    @mark
    def test_open_scaled_in_update_mode(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/119
        (Don't update scaled image data if the data is not read)

        This ensures that merely opening and closing a file containing scaled
        image data does not cause any change to the data (or the header).
        Changes should only occur if the data is accessed.
        """

        # Copy the original file before making any possible changes to it
        self.copy_file("scale.fits")
        mtime = os.stat(self.temp("scale.fits")).st_mtime

        time.sleep(1)

        fits.open(self.temp("scale.fits"), mode="update").close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert mtime == os.stat(self.temp("scale.fits")).st_mtime

        # Insert a slight delay to ensure the mtime does change when the file
        # is changed
        time.sleep(1)

        hdul = fits.open(self.temp("scale.fits"), "update")
        orig_data = hdul[0].data
        hdul.close()

        # Now the file should be updated with the rescaled data
        assert mtime != os.stat(self.temp("scale.fits")).st_mtime
        hdul = fits.open(self.temp("scale.fits"), mode="update")
        assert hdul[0].data.dtype == np.dtype(">f4")
        assert hdul[0].header["BITPIX"] == -32
        assert "BZERO" not in hdul[0].header
        assert "BSCALE" not in hdul[0].header
        assert (orig_data == hdul[0].data).all()

        # Try reshaping the data, then closing and reopening the file; let's
        # see if all the changes are preserved properly
        hdul[0].data.shape = (42, 10)
        hdul.close()

        hdul = fits.open(self.temp("scale.fits"))
        assert hdul[0].shape == (42, 10)
        assert hdul[0].data.dtype == np.dtype(">f4")
        assert hdul[0].header["BITPIX"] == -32
        assert "BZERO" not in hdul[0].header
        assert "BSCALE" not in hdul[0].header
        hdul.close()

    def test_scale_back(self):
        """A simple test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/120

        The scale_back feature for image HDUs.
        """

        self.copy_file("scale.fits")
        with fits.open(self.temp("scale.fits"), mode="update", scale_back=True) as hdul:
            orig_bitpix = hdul[0].header["BITPIX"]
            orig_bzero = hdul[0].header["BZERO"]
            orig_bscale = hdul[0].header["BSCALE"]
            orig_data = hdul[0].data.copy()
            hdul[0].data[0] = 0

        with fits.open(self.temp("scale.fits"), do_not_scale_image_data=True) as hdul:
            assert hdul[0].header["BITPIX"] == orig_bitpix
            assert hdul[0].header["BZERO"] == orig_bzero
            assert hdul[0].header["BSCALE"] == orig_bscale

            zero_point = int(math.floor(-orig_bzero / orig_bscale))
            assert (hdul[0].data[0] == zero_point).all()

        with fits.open(self.temp("scale.fits")) as hdul:
            assert (hdul[0].data[1:] == orig_data[1:]).all()

    def test_image_none(self):
        """
        Regression test for https://github.com/spacetelescope/PyFITS/issues/27
        """

        with fits.open(self.data("test0.fits")) as h:
            h[1].data
            h[1].data = None
            h[1].writeto(self.temp("test.fits"))

        with fits.open(self.temp("test.fits")) as h:
            assert h[1].data is None
            assert h[1].header["NAXIS"] == 0
            assert "NAXIS1" not in h[1].header
            assert "NAXIS2" not in h[1].header

    def test_invalid_blank(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2711

        If the BLANK keyword contains an invalid value it should be ignored for
        any calculations (though a warning should be issued).
        """

        data = np.arange(100, dtype=np.float64)
        hdu = fits.PrimaryHDU(data)
        hdu.header["BLANK"] = "nan"

        with pytest.warns(fits.verify.VerifyWarning) as w:
            hdu.writeto(self.temp("test.fits"))
        assert "Invalid value for 'BLANK' keyword in header: 'nan'" in str(w[0].message)

        with pytest.warns(AstropyUserWarning) as w:
            with fits.open(self.temp("test.fits")) as hdul:
                assert np.all(hdul[0].data == data)

        assert len(w) == 2
        msg = "Invalid value for 'BLANK' keyword in header"
        assert msg in str(w[0].message)
        msg = "Invalid 'BLANK' keyword"
        assert msg in str(w[1].message)

    def test_scaled_image_fromfile(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2710
        """

        # Make some sample data
        a = np.arange(100, dtype=np.float32)

        hdu = fits.PrimaryHDU(data=a.copy())
        hdu.scale(bscale=1.1)
        hdu.writeto(self.temp("test.fits"))

        with open(self.temp("test.fits"), "rb") as f:
            file_data = f.read()

        hdul = fits.HDUList.fromstring(file_data)
        assert np.allclose(hdul[0].data, a)

    def test_set_data(self):
        """
        Test data assignment - issue #5087
        """

        im = fits.ImageHDU()
        ar = np.arange(12)
        im.data = ar

    def test_scale_bzero_with_int_data(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/4600
        """

        a = np.arange(100, 200, dtype=np.int16)

        hdu1 = fits.PrimaryHDU(data=a.copy())
        hdu2 = fits.PrimaryHDU(data=a.copy())
        # Previously the following line would throw a TypeError,
        # now it should be identical to the integer bzero case
        hdu1.scale("int16", bzero=99.0)
        hdu2.scale("int16", bzero=99)
        assert np.allclose(hdu1.data, hdu2.data)

    def test_scale_back_uint_assignment(self):
        """
        Extend fix for #4600 to assignment to data

        Suggested by:
        https://github.com/astropy/astropy/pull/4602#issuecomment-208713748
        """

        a = np.arange(100, 200, dtype=np.uint16)
        fits.PrimaryHDU(a).writeto(self.temp("test.fits"))
        with fits.open(self.temp("test.fits"), mode="update", scale_back=True) as (
            hdu,
        ):
            hdu.data[:] = 0
            assert np.allclose(hdu.data, 0)

    def test_hdu_creation_with_scalar(self):
        msg = r"data object array\(1\) should have at least one dimension"
        with pytest.raises(TypeError, match=msg):
            fits.ImageHDU(data=1)
        with pytest.raises(TypeError, match=msg):
            fits.PrimaryHDU(data=1)
        # Regression test for https://github.com/astropy/astropy/issues/14527
        with pytest.raises(TypeError, match=msg):
            fits.ImageHDU(data=np.array(1))
        with pytest.raises(TypeError, match=msg):
            fits.PrimaryHDU(data=np.array(1))


def test_scale_implicit_casting():
    # Regression test for an issue that occurred because Numpy now does not
    # allow implicit type casting during inplace operations.

    hdu = fits.ImageHDU(np.array([1], dtype=np.int32))
    hdu.scale(bzero=1.3)


def test_bzero_implicit_casting_compressed():
    # Regression test for an issue that occurred because Numpy now does not
    # allow implicit type casting during inplace operations. Astropy is
    # actually not able to produce a file that triggers the failure - the
    # issue occurs when using unsigned integer types in the FITS file, in which
    # case BZERO should be 32768. But if the keyword is stored as 32768.0, then
    # it was possible to trigger the implicit casting error.

    filename = get_pkg_data_filename("data/compressed_float_bzero.fits")

    with fits.open(filename) as hdul:
        hdu = hdul[1]
        hdu.data


def test_bzero_mishandled_info(tmp_path):
    # Regression test for #5507:
    # Calling HDUList.info() on a dataset which applies a zeropoint
    # from BZERO but which astropy.io.fits does not think it needs
    # to resize to a new dtype results in an AttributeError.
    filename = tmp_path / "floatimg_with_bzero.fits"
    hdu = fits.ImageHDU(np.zeros((10, 10)))
    hdu.header["BZERO"] = 10
    hdu.writeto(filename, overwrite=True)
    with fits.open(filename) as hdul:
        hdul.info()


def test_image_write_readonly(tmp_path):
    # Regression test to make sure that we can write out read-only arrays (#5512)

    x = np.array([1, 2, 3])
    x.setflags(write=False)
    ghdu = fits.ImageHDU(data=x)
    ghdu.add_datasum()

    filename = tmp_path / "test.fits"

    ghdu.writeto(filename)

    with fits.open(filename) as hdulist:
        assert_equal(hdulist[1].data, [1, 2, 3])


def test_int8(tmp_path):
    """Test for int8 support, https://github.com/astropy/astropy/issues/11995"""
    img = np.arange(-50, 50, dtype=np.int8).reshape(10, 10)
    hdu = fits.PrimaryHDU(img)
    hdu.writeto(tmp_path / "int8.fits")

    with fits.open(tmp_path / "int8.fits") as hdul:
        assert hdul[0].header["BITPIX"] == 8
        assert hdul[0].header["BZERO"] == -128
        assert hdul[0].header["BSCALE"] == 1.0
        assert_equal(hdul[0].data, img)
        assert hdul[0].data.dtype == img.dtype



#cython: language_level=3
cimport cython
cimport numpy as np
from libc cimport math

np.import_array()

ctypedef fused dtype:
    np.uint8_t
    np.uint16_t
    np.uint32_t
    np.uint64_t
    np.int8_t
    np.int16_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t
    np.longdouble_t


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef double ks_2samp(dtype[:] data1, dtype[:] data2):
    cdef:
        size_t i = 0, j = 0, n1 = data1.shape[0], n2 = data2.shape[0]
        dtype d1i, d2j
        double d = 0, mind = 0, maxd = 0, inv_n1 = 1. / n1, inv_n2 = 1. / n2
    while i < n1 and j < n2:
        d1i = data1[i]
        d2j = data2[j]
        if d1i <= d2j:
            while i < n1 and data1[i] == d1i:
                d += inv_n1
                i += 1
        if d1i >= d2j:
            while j < n2 and data2[j] == d2j:
                d -= inv_n2
                j += 1
        mind = min(mind, d)
        maxd = max(maxd, d)
    return maxd - mind



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides the tools used to internally run the astropy test suite
from the installed astropy.  It makes use of the |pytest| testing framework.
"""

import os
import pickle
import sys

import pytest

from astropy.units import allclose as quantity_allclose  # noqa: F401
from astropy.utils.introspection import minversion

# For backward-compatibility with affiliated packages
from .runner import TestRunner  # noqa: F401

__all__ = [
    "assert_follows_unicode_guidelines",
    "assert_quantity_allclose",
    "check_pickling_recovery",
    "pickle_protocol",
    "generic_recursive_equality_test",
]

PYTEST_LT_8_0 = not minversion(pytest, "8.0")

# https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
CI = os.environ.get("CI", "false") == "true"
IS_CRON = os.environ.get("IS_CRON", "false") == "true"


def _save_coverage(cov, result, rootdir, testing_path):
    """
    This method is called after the tests have been run in coverage mode
    to cleanup and then save the coverage data and report.
    """
    from astropy.utils.console import color_print

    if result != 0:
        return

    # The coverage report includes the full path to the temporary
    # directory, so we replace all the paths with the true source
    # path. Note that this will not work properly for packages that still
    # rely on 2to3.
    try:
        # Coverage 4.0: _harvest_data has been renamed to get_data, the
        # lines dict is private
        cov.get_data()
    except AttributeError:
        # Coverage < 4.0
        cov._harvest_data()
        lines = cov.data.lines
    else:
        lines = cov.data._lines

    for key in list(lines.keys()):
        new_path = os.path.relpath(
            os.path.realpath(key), os.path.realpath(testing_path)
        )
        new_path = os.path.abspath(os.path.join(rootdir, new_path))
        lines[new_path] = lines.pop(key)

    color_print("Saving coverage data in .coverage...", "green")
    cov.save()

    color_print("Saving HTML coverage report in htmlcov...", "green")
    cov.html_report(directory=os.path.join(rootdir, "htmlcov"))


def assert_follows_unicode_guidelines(x, roundtrip=None):
    """
    Test that an object follows our Unicode policy.  See
    "Unicode guidelines" in the coding guidelines.

    Parameters
    ----------
    x : object
        The instance to test

    roundtrip : module, optional
        When provided, this namespace will be used to evaluate
        ``repr(x)`` and ensure that it roundtrips.  It will also
        ensure that ``__bytes__(x)`` roundtrip.
        If not provided, no roundtrip testing will be performed.
    """
    from astropy import conf

    with conf.set_temp("unicode_output", False):
        bytes_x = bytes(x)
        unicode_x = str(x)
        repr_x = repr(x)

        assert isinstance(bytes_x, bytes)
        bytes_x.decode("ascii")
        assert isinstance(unicode_x, str)
        unicode_x.encode("ascii")
        assert isinstance(repr_x, str)
        if isinstance(repr_x, bytes):
            repr_x.decode("ascii")
        else:
            repr_x.encode("ascii")

        if roundtrip is not None:
            assert x.__class__(bytes_x) == x
            assert x.__class__(unicode_x) == x
            assert eval(repr_x, roundtrip) == x

    with conf.set_temp("unicode_output", True):
        bytes_x = bytes(x)
        unicode_x = str(x)
        repr_x = repr(x)

        assert isinstance(bytes_x, bytes)
        bytes_x.decode("ascii")
        assert isinstance(unicode_x, str)
        assert isinstance(repr_x, str)
        if isinstance(repr_x, bytes):
            repr_x.decode("ascii")
        else:
            repr_x.encode("ascii")

        if roundtrip is not None:
            assert x.__class__(bytes_x) == x
            assert x.__class__(unicode_x) == x
            assert eval(repr_x, roundtrip) == x


@pytest.fixture(params=[0, 1, -1])
def pickle_protocol(request):
    """
    Fixture to run all the tests for protocols 0 and 1, and -1 (most advanced).
    (Originally from astropy.table.tests.test_pickle).
    """
    return request.param


def generic_recursive_equality_test(a, b, class_history):
    """
    Check if the attributes of a and b are equal. Then,
    check if the attributes of the attributes are equal.
    """
    if sys.version_info < (3, 11):
        dict_a = a.__getstate__() if hasattr(a, "__getstate__") else a.__dict__
    else:
        # NOTE: The call may need to be adapted if other objects implementing a __getstate__
        # with required argument(s) are passed to this function.
        # For a class with `__slots__` the default state is not a `dict`;
        # with neither `__dict__` nor `__slots__` it is `None`.
        state = a.__getstate__(a) if isinstance(a, type) else a.__getstate__()
        dict_a = state if isinstance(state, dict) else getattr(a, "__dict__", dict())
    dict_b = b.__dict__
    for key in dict_a:
        assert key in dict_b, f"Did not pickle {key}"

        if dict_a[key].__class__.__eq__ is not object.__eq__:
            # Only compare if the class defines a proper equality test.
            # E.g., info does not define __eq__, and hence defers to
            # object.__eq__, which is equivalent to checking that two
            # instances are the same.  This will generally not be true
            # after pickling.
            eq = dict_a[key] == dict_b[key]
            if "__iter__" in dir(eq):
                eq = False not in eq
            assert eq, f"Value of {key} changed by pickling"

        if hasattr(dict_a[key], "__dict__"):
            if dict_a[key].__class__ in class_history:
                # attempt to prevent infinite recursion
                pass
            else:
                new_class_history = [dict_a[key].__class__]
                new_class_history.extend(class_history)
                generic_recursive_equality_test(
                    dict_a[key], dict_b[key], new_class_history
                )


def check_pickling_recovery(original, protocol):
    """
    Try to pickle an object. If successful, make sure
    the object's attributes survived pickling and unpickling.
    """
    f = pickle.dumps(original, protocol=protocol)
    unpickled = pickle.loads(f)
    class_history = [original.__class__]
    generic_recursive_equality_test(original, unpickled, class_history)


def assert_quantity_allclose(actual, desired, rtol=1.0e-7, atol=None, **kwargs):
    """
    Raise an assertion if two objects are not equal up to desired tolerance.

    This is a :class:`~astropy.units.Quantity`-aware version of
    :func:`numpy.testing.assert_allclose`.
    """
    import numpy as np

    from astropy.units.quantity import _unquantify_allclose_arguments

    np.testing.assert_allclose(
        *_unquantify_allclose_arguments(actual, desired, rtol, atol), **kwargs
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This file is the main file used when running tests with pytest directly,
# in particular if running e.g. ``pytest docs/``.

import os
import tempfile

import hypothesis

from astropy import __version__

try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
except ImportError:
    PYTEST_HEADER_MODULES = {}
    TESTED_VERSIONS = {}


# This has to be in the root dir or it will not display in CI.
def pytest_configure(config):
    PYTEST_HEADER_MODULES["PyERFA"] = "erfa"
    PYTEST_HEADER_MODULES["Cython"] = "cython"
    PYTEST_HEADER_MODULES["Scikit-image"] = "skimage"
    PYTEST_HEADER_MODULES["pyarrow"] = "pyarrow"
    PYTEST_HEADER_MODULES["asdf-astropy"] = "asdf_astropy"
    TESTED_VERSIONS["Astropy"] = __version__


# This has to be in the root dir or it will not display in CI.
def pytest_report_header(config):
    # This gets added after the pytest-astropy-header output.
    return (
        f'CI: {os.environ.get("CI", "undefined")}\n'
        f'ARCH_ON_CI: {os.environ.get("ARCH_ON_CI", "undefined")}\n'
        f'IS_CRON: {os.environ.get("IS_CRON", "undefined")}\n'
    )


# Tell Hypothesis that we might be running slow tests, to print the seed blob
# so we can easily reproduce failures from CI, and derive a fuzzing profile
# to try many more inputs when we detect a scheduled build or when specifically
# requested using the HYPOTHESIS_PROFILE=fuzz environment variable or
# `pytest --hypothesis-profile=fuzz ...` argument.

hypothesis.settings.register_profile(
    "ci",
    deadline=None,
    print_blob=True,
    derandomize=True,
    # disabling HealthCheck.differing_executors to allow double test
    # see https://github.com/astropy/astropy/issues/17299
    suppress_health_check=[hypothesis.HealthCheck.differing_executors],
)
hypothesis.settings.register_profile(
    "fuzzing", deadline=None, print_blob=True, max_examples=1000
)
default = (
    "fuzzing"
    if (
        os.environ.get("IS_CRON") == "true"
        and os.environ.get("ARCH_ON_CI") not in ("aarch64", "ppc64le")
    )
    else "ci"
)
hypothesis.settings.load_profile(os.environ.get("HYPOTHESIS_PROFILE", default))

# Make sure we use temporary directories for the config and cache
# so that the tests are insensitive to local configuration.

os.environ["XDG_CONFIG_HOME"] = tempfile.mkdtemp("astropy_config")
os.environ["XDG_CACHE_HOME"] = tempfile.mkdtemp("astropy_cache")

os.mkdir(os.path.join(os.environ["XDG_CONFIG_HOME"], "astropy"))
os.mkdir(os.path.join(os.environ["XDG_CACHE_HOME"], "astropy"))

# Note that we don't need to change the environment variables back or remove
# them after testing, because they are only changed for the duration of the
# Python process, and this configuration only matters if running pytest
# directly, not from e.g. an IPython session.



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import numpy as np

from astropy.io import fits
from astropy.io.fits.tests.test_checksum import BaseChecksumTests


class TestChecksumFunctions(BaseChecksumTests):
    # All checksums have been verified against CFITSIO

    def test_compressed_image_data(self):
        with fits.open(self.data("comp.fits")) as h1:
            h1.writeto(self.temp("tmp.fits"), overwrite=True, checksum=True)
            with fits.open(self.temp("tmp.fits"), checksum=True) as h2:
                assert np.all(h1[1].data == h2[1].data)
                assert "CHECKSUM" in h2[0].header
                assert h2[0].header["CHECKSUM"] == "D8iBD6ZAD6fAD6ZA"
                assert "DATASUM" in h2[0].header
                assert h2[0].header["DATASUM"] == "0"
                assert "CHECKSUM" in h2[1].header
                assert h2[1].header["CHECKSUM"] == "ZeAbdb8aZbAabb7a"
                assert "DATASUM" in h2[1].header
                assert h2[1].header["DATASUM"] == "113055149"

    def test_failing_compressed_datasum(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/4587
        """
        n = np.ones((10, 10), dtype="float32")
        comp_hdu = fits.CompImageHDU(n)
        comp_hdu.writeto(self.temp("tmp.fits"), checksum=True)

        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert np.all(hdul[1].data == comp_hdu.data)

    def test_compressed_image_data_int16(self):
        n = np.arange(100, dtype="int16")
        hdu = fits.ImageHDU(n)
        comp_hdu = fits.CompImageHDU(hdu.data, hdu.header)
        comp_hdu.writeto(self.temp("tmp.fits"), checksum=True)
        hdu.writeto(self.temp("uncomp.fits"), checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert np.all(hdul[1].data == comp_hdu.data)
            assert np.all(hdul[1].data == hdu.data)
            assert "CHECKSUM" in hdul[0].header
            assert hdul[0].header["CHECKSUM"] == "D8iBD6ZAD6fAD6ZA"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "0"

            assert "CHECKSUM" in hdul[1].header
            assert hdul[1]._header["CHECKSUM"] == "J5cCJ5c9J5cAJ5c9"
            assert "DATASUM" in hdul[1].header
            assert hdul[1]._header["DATASUM"] == "2453673070"
            assert "CHECKSUM" in hdul[1].header

            with fits.open(self.temp("uncomp.fits"), checksum=True) as hdul2:
                header_comp = hdul[1]._header
                header_uncomp = hdul2[1].header
                assert "ZHECKSUM" in header_comp
                assert "CHECKSUM" in header_uncomp
                assert header_uncomp["CHECKSUM"] == "ZE94eE91ZE91bE91"
                assert header_comp["ZHECKSUM"] == header_uncomp["CHECKSUM"]
                assert "ZDATASUM" in header_comp
                assert "DATASUM" in header_uncomp
                assert header_uncomp["DATASUM"] == "160565700"
                assert header_comp["ZDATASUM"] == header_uncomp["DATASUM"]

    def test_compressed_image_data_float32(self):
        n = np.arange(100, dtype="float32")
        hdu = fits.ImageHDU(n)
        comp_hdu = fits.CompImageHDU(hdu.data, hdu.header)
        comp_hdu.writeto(self.temp("tmp.fits"), checksum=True)
        hdu.writeto(self.temp("uncomp.fits"), checksum=True)
        with fits.open(self.temp("tmp.fits"), checksum=True) as hdul:
            assert np.all(hdul[1].data == comp_hdu.data)
            assert np.all(hdul[1].data == hdu.data)
            assert "CHECKSUM" in hdul[0].header
            assert hdul[0].header["CHECKSUM"] == "D8iBD6ZAD6fAD6ZA"
            assert "DATASUM" in hdul[0].header
            assert hdul[0].header["DATASUM"] == "0"

            assert "CHECKSUM" in hdul[1].header
            assert "DATASUM" in hdul[1].header

            # The checksum ends up being different on Windows and s390/bigendian,
            # possibly due to slight floating point differences? See gh-10921.
            # TODO fix these so they work on all platforms; otherwise pointless.
            # assert hdul[1]._header['CHECKSUM'] == 'eATIf3SHe9SHe9SH'
            # assert hdul[1]._header['DATASUM'] == '1277667818'

            with fits.open(self.temp("uncomp.fits"), checksum=True) as hdul2:
                header_comp = hdul[1]._header
                header_uncomp = hdul2[1].header
                assert "ZHECKSUM" in header_comp
                assert "CHECKSUM" in header_uncomp
                assert header_uncomp["CHECKSUM"] == "Cgr5FZo2Cdo2CZo2"
                assert header_comp["ZHECKSUM"] == header_uncomp["CHECKSUM"]
                assert "ZDATASUM" in header_comp
                assert "DATASUM" in header_uncomp
                assert header_uncomp["DATASUM"] == "2393636889"
                assert header_comp["ZDATASUM"] == header_uncomp["DATASUM"]



import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.coordinates import SkyCoord
from astropy.wcs.wcsapi.high_level_wcs_wrapper import HighLevelWCSWrapper
from astropy.wcs.wcsapi.low_level_api import BaseLowLevelWCS


class CustomLowLevelWCS(BaseLowLevelWCS):
    @property
    def pixel_n_dim(self):
        return 2

    @property
    def world_n_dim(self):
        return 2

    @property
    def world_axis_physical_types(self):
        return ["pos.eq.ra", "pos.eq.dec"]

    @property
    def world_axis_units(self):
        return ["deg", "deg"]

    def pixel_to_world_values(self, *pixel_arrays):
        return [np.asarray(pix) * 2 for pix in pixel_arrays]

    def world_to_pixel_values(self, *world_arrays):
        return [np.asarray(world) / 2 for world in world_arrays]

    @property
    def world_axis_object_components(self):
        return [
            ("test", 0, "spherical.lon.degree"),
            ("test", 1, "spherical.lat.degree"),
        ]

    @property
    def world_axis_object_classes(self):
        return {"test": (SkyCoord, (), {"unit": "deg"})}


def test_wrapper():
    wcs = CustomLowLevelWCS()

    wrapper = HighLevelWCSWrapper(wcs)

    coord = wrapper.pixel_to_world(1, 2)

    assert isinstance(coord, SkyCoord)
    assert coord.isscalar

    x, y = wrapper.world_to_pixel(coord)

    assert_allclose(x, 1)
    assert_allclose(y, 2)

    assert wrapper.low_level_wcs is wcs
    assert wrapper.pixel_n_dim == 2
    assert wrapper.world_n_dim == 2
    assert wrapper.world_axis_physical_types == ["pos.eq.ra", "pos.eq.dec"]
    assert wrapper.world_axis_units == ["deg", "deg"]
    assert wrapper.array_shape is None
    assert wrapper.pixel_bounds is None
    assert np.all(wrapper.axis_correlation_matrix)


def test_wrapper_invalid():
    class InvalidCustomLowLevelWCS(CustomLowLevelWCS):
        @property
        def world_axis_object_classes(self):
            return {}

    wcs = InvalidCustomLowLevelWCS()

    wrapper = HighLevelWCSWrapper(wcs)

    with pytest.raises(KeyError):
        wrapper.pixel_to_world(1, 2)



# Create a planetary WCS structure

from astropy import units as u
from astropy.coordinates import BaseBodycentricRepresentation, BaseCoordinateFrame
from astropy.wcs.utils import celestial_frame_to_wcs


class MARSCustomBodycentricRepresentation(BaseBodycentricRepresentation):
    _equatorial_radius = 3399190.0 * u.m
    _flattening = 0.5886 * u.percent


class MARSCustomBodyFrame(BaseCoordinateFrame):
    name = "Mars"


frame = MARSCustomBodyFrame()
frame.representation_type = MARSCustomBodycentricRepresentation
mywcs = celestial_frame_to_wcs(frame, projection="CAR")
print(mywcs.wcs.ctype)
print(mywcs.wcs.name)
print(mywcs.wcs.aux.a_radius)
print(mywcs.wcs.aux.c_radius)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import numpy as np
import pytest

from astropy.nddata import FlagCollection


def test_init():
    FlagCollection(shape=(1, 2, 3))


def test_init_noshape():
    with pytest.raises(Exception) as exc:
        FlagCollection()
    assert (
        exc.value.args[0]
        == "FlagCollection should be initialized with the shape of the data"
    )


def test_init_notiterable():
    with pytest.raises(Exception) as exc:
        FlagCollection(shape=1.0)
    assert exc.value.args[0] == "FlagCollection shape should be an iterable object"


def test_setitem():
    f = FlagCollection(shape=(1, 2, 3))
    f["a"] = np.ones((1, 2, 3)).astype(float)
    f["b"] = np.ones((1, 2, 3)).astype(int)
    f["c"] = np.ones((1, 2, 3)).astype(bool)
    f["d"] = np.ones((1, 2, 3)).astype(str)


@pytest.mark.parametrize("value", [1, 1.0, "spam", [1, 2, 3], (1.0, 2.0, 3.0)])
def test_setitem_invalid_type(value):
    f = FlagCollection(shape=(1, 2, 3))
    with pytest.raises(Exception) as exc:
        f["a"] = value
    assert exc.value.args[0] == "flags should be given as a Numpy array"


def test_setitem_invalid_shape():
    f = FlagCollection(shape=(1, 2, 3))
    with pytest.raises(ValueError) as exc:
        f["a"] = np.ones((3, 2, 1))
    assert exc.value.args[0].startswith("flags array shape")
    assert exc.value.args[0].endswith("does not match data shape (1, 2, 3)")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.flrw.wpwazpcdm`."""

import numpy as np
import pytest

import astropy.cosmology.units as cu
import astropy.units as u
from astropy.cosmology import FlatwpwaCDM, wpwaCDM
from astropy.cosmology.parameter import Parameter
from astropy.cosmology.tests.test_core import ParameterTestMixin
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.compat.optional_deps import HAS_SCIPY

from .conftest import filter_keys_from_items
from .test_base import FlatFLRWMixinTest, FLRWTest
from .test_w0wacdm import ParameterwaTestMixin

##############################################################################
# PARAMETERS

COMOVING_DISTANCE_EXAMPLE_KWARGS = {"wp": -0.9, "zp": 0.5, "wa": 0.1, "Tcmb0": 0.0}


##############################################################################
# TESTS
##############################################################################


class ParameterwpTestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` wp on a Cosmology.

    wp is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_wp(self, cosmo_cls, cosmo):
        """Test Parameter ``wp``."""
        # on the class
        wp = cosmo_cls.parameters["wp"]
        assert isinstance(wp, Parameter)
        assert "at the pivot" in wp.__doc__
        assert wp.unit is None
        assert wp.default == -1.0

        # on the instance
        assert cosmo.wp is cosmo._wp
        assert cosmo.wp == self.cls_kwargs["wp"]

    def test_init_wp(self, cosmo_cls, ba):
        """Test initialization for values of ``wp``."""
        # test that it works with units
        ba.arguments["wp"] = ba.arguments["wp"] << u.one  # ensure units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.wp == ba.arguments["wp"]

        # also without units
        ba.arguments["wp"] = ba.arguments["wp"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.wp == ba.arguments["wp"]

        # must be dimensionless
        ba.arguments["wp"] = 10 * u.km
        with pytest.raises(TypeError):
            cosmo_cls(*ba.args, **ba.kwargs)


class ParameterzpTestMixin(ParameterTestMixin):
    """Tests for `astropy.cosmology.Parameter` zp on a Cosmology.

    zp is a descriptor, which are tested by mixin, here with ``TestFLRW``.
    These tests expect dicts ``_cls_args`` and ``cls_kwargs`` which give the
    args and kwargs for the cosmology class, respectively. See ``TestFLRW``.
    """

    def test_zp(self, cosmo_cls, cosmo):
        """Test Parameter ``zp``."""
        # on the class
        zp = cosmo_cls.parameters["zp"]
        assert isinstance(zp, Parameter)
        assert "pivot redshift" in zp.__doc__
        assert zp.unit == cu.redshift
        assert zp.default == 0.0

        # on the instance
        assert cosmo.zp is cosmo._zp
        assert cosmo.zp == self.cls_kwargs["zp"] << cu.redshift

    def test_init_zp(self, cosmo_cls, ba):
        """Test initialization for values of ``zp``."""
        # test that it works with units
        ba.arguments["zp"] = ba.arguments["zp"] << u.one  # ensure units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.zp == ba.arguments["zp"]

        # also without units
        ba.arguments["zp"] = ba.arguments["zp"].value  # strip units
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)
        assert cosmo.zp.value == ba.arguments["zp"]

        # must be dimensionless
        ba.arguments["zp"] = 10 * u.km
        with pytest.raises(u.UnitConversionError):
            cosmo_cls(*ba.args, **ba.kwargs)


class TestwpwaCDM(
    FLRWTest, ParameterwpTestMixin, ParameterwaTestMixin, ParameterzpTestMixin
):
    """Test :class:`astropy.cosmology.wpwaCDM`."""

    def setup_class(self):
        """Setup for testing."""
        super().setup_class(self)
        self.cls = wpwaCDM
        self.cls_kwargs.update(wp=-0.9, wa=0.2, zp=0.5)

    # ===============================================================
    # Method & Attribute Tests

    def test_clone_change_param(self, cosmo):
        """Test method ``.clone()`` changing a(many) Parameter(s)."""
        super().test_clone_change_param(cosmo)

        # `w` params
        c = cosmo.clone(wp=0.1, wa=0.2, zp=14)
        assert c.wp == 0.1
        assert c.wa == 0.2
        assert c.zp == 14
        for n, v in filter_keys_from_items(c.parameters, ("wp", "wa", "zp")):
            v_expect = getattr(cosmo, n)
            if v is None:
                assert v is v_expect
            else:
                assert_quantity_allclose(v, v_expect, atol=1e-4 * getattr(v, "unit", 1))

    # @pytest.mark.parametrize("z", valid_zs)  # TODO! recompute comparisons below
    def test_w(self, cosmo):
        """Test :meth:`astropy.cosmology.wpwaCDM.w`."""
        # super().test_w(cosmo, z)

        assert u.allclose(cosmo.w(0.5), -0.9)
        assert u.allclose(
            cosmo.w([0.1, 0.2, 0.5, 1.5, 2.5, 11.5]),
            [-0.94848485, -0.93333333, -0.9, -0.84666667, -0.82380952, -0.78266667],
        )

    def test_repr(self, cosmo_cls, cosmo):
        """Test method ``.__repr__()``."""
        assert repr(cosmo) == (
            "wpwaCDM(name='ABCMeta', H0=<Quantity 70. km / (Mpc s)>, Om0=0.27,"
            " Ode0=0.73, Tcmb0=<Quantity 3. K>, Neff=3.04,"
            " m_nu=<Quantity [0., 0., 0.] eV>, Ob0=0.03, wp=-0.9, wa=0.2,"
            " zp=<Quantity 0.5 redshift>)"
        )

    # ===============================================================
    # Usage Tests

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy required for this test.")
    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [
            (  # no relativistic species
                (75.0, 0.3, 0.6),
                {},
                [2954.68975298, 4599.83254834, 5643.04013201, 6373.36147627] * u.Mpc,
            ),
            (  # massless neutrinos
                (75.0, 0.25, 0.5),
                {"zp": 0.4, "Tcmb0": 3.0, "Neff": 3, "m_nu": 0 * u.eV},
                [2919.00656215, 4558.0218123, 5615.73412391, 6366.10224229] * u.Mpc,
            ),
            (  # massive neutrinos
                (75.0, 0.25, 0.5),
                {"zp": 1.0, "Tcmb0": 3.0, "Neff": 4, "m_nu": 5 * u.eV},
                [2629.48489827, 3874.13392319, 4614.31562397, 5116.51184842] * u.Mpc,
            ),
            # FLAT: these match the tests in TestFlatwpwaCDM, except Ode0 is set manually.
            (  # no relativistic species
                (75.0, 0.3, 0.7),
                {},
                [3030.70481348, 4745.82435272, 5828.73710847, 6582.60454542] * u.Mpc,
            ),
            (  # massless neutrinos
                (75.0, 0.25, 0.75),
                {"zp": 0.4, "Tcmb0": 3.0, "Neff": 3, "m_nu": 0 * u.eV},
                [3113.62199365, 4943.28425668, 6114.45491003, 6934.07461377] * u.Mpc,
            ),
            (  # massive neutrinos
                (75.0, 0.25, 0.2458794183661),  # to make Ok0 = 0, Otot0 = 1
                {"zp": 1.0, "Tcmb0": 3.0, "Neff": 4, "m_nu": 5 * u.eV},
                [2517.08634022, 3694.21111754, 4402.17802962, 4886.65787948] * u.Mpc,
            ),
        ],
    )
    def test_comoving_distance_example(self, cosmo_cls, args, kwargs, expected):
        """Test :meth:`astropy.cosmology.LambdaCDM.comoving_distance`.

        These do not come from external codes -- they are just internal checks to make
        sure nothing changes if we muck with the distance calculators.
        """
        super().test_comoving_distance_example(
            cosmo_cls, args, {**COMOVING_DISTANCE_EXAMPLE_KWARGS, **kwargs}, expected
        )


class TestFlatwpwaCDM(FlatFLRWMixinTest, TestwpwaCDM):
    """Test :class:`astropy.cosmology.FlatwpwaCDM`."""

    def setup_class(self):
        """Setup for testing."""
        super().setup_class(self)
        self.cls = FlatwpwaCDM

    def test_repr(self, cosmo_cls, cosmo):
        """Test method ``.__repr__()``."""
        super().test_repr(cosmo_cls, cosmo)

        assert repr(cosmo) == (
            "FlatwpwaCDM(name='ABCMeta', H0=<Quantity 70. km / (Mpc s)>, Om0=0.27,"
            " Tcmb0=<Quantity 3. K>, Neff=3.04, m_nu=<Quantity [0., 0., 0.] eV>,"
            " Ob0=0.03, wp=-0.9, wa=0.2, zp=<Quantity 0.5 redshift>)"
        )

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy required for this test.")
    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [
            (  # no relativistic species
                (75.0, 0.3),
                {},
                [3030.70481348, 4745.82435272, 5828.73710847, 6582.60454542] * u.Mpc,
            ),
            (  # massless neutrinos
                (75.0, 0.25),
                {"zp": 0.4, "wa": 0.1, "Tcmb0": 3.0, "Neff": 3, "m_nu": 0.0 * u.eV},
                [3113.62199365, 4943.28425668, 6114.45491003, 6934.07461377] * u.Mpc,
            ),
            (  # massive neutrinos
                (75.0, 0.25),
                {"zp": 1.0, "Tcmb0": 3.0, "Neff": 4, "m_nu": 5 * u.eV},
                [2517.08634022, 3694.21111754, 4402.17802962, 4886.65787948] * u.Mpc,
            ),
        ],
    )
    def test_comoving_distance_example(self, cosmo_cls, args, kwargs, expected):
        """Test :meth:`astropy.cosmology.LambdaCDM.comoving_distance`.

        These do not come from external codes -- they are just internal checks to make
        sure nothing changes if we muck with the distance calculators.
        """
        super().test_comoving_distance_example(
            cosmo_cls, args, {**COMOVING_DISTANCE_EXAMPLE_KWARGS, **kwargs}, expected
        )


###############################################################################
# Comparison to Other Codes


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy.")
def test_varyde_lumdist_mathematica():
    """Tests a few varying dark energy EOS models against a Mathematica computation."""
    z = np.array([0.2, 0.4, 0.9, 1.2])

    # wpwa models
    cosmo = wpwaCDM(H0=70, Om0=0.2, Ode0=0.8, wp=-1.1, wa=0.2, zp=0.5, Tcmb0=0.0)
    assert u.allclose(
        cosmo.luminosity_distance(z),
        [1010.81, 2294.45, 6369.45, 9218.95] * u.Mpc,
        rtol=1e-4,
    )

    cosmo = wpwaCDM(H0=70, Om0=0.2, Ode0=0.8, wp=-1.1, wa=0.2, zp=0.9, Tcmb0=0.0)
    assert u.allclose(
        cosmo.luminosity_distance(z),
        [1013.68, 2305.3, 6412.37, 9283.33] * u.Mpc,
        rtol=1e-4,
    )


##############################################################################
# Miscellaneous
# TODO: these should be better integrated into the new test framework


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
def test_de_densityscale():
    cosmo = wpwaCDM(H0=70, Om0=0.3, Ode0=0.70, wp=-0.9, wa=0.2, zp=0.5)

    z = np.array([0.1, 0.2, 0.5, 1.5, 2.5])
    assert u.allclose(
        cosmo.de_density_scale(z),
        [1.012246048, 1.0280102, 1.087439, 1.324988, 1.565746],
        rtol=1e-4,
    )

    assert u.allclose(cosmo.de_density_scale(3), cosmo.de_density_scale(3.0), rtol=1e-7)
    assert u.allclose(
        cosmo.de_density_scale([1, 2, 3]),
        cosmo.de_density_scale([1.0, 2.0, 3.0]),
        rtol=1e-7,
    )

    # Flat tests
    cosmo = wpwaCDM(H0=70, Om0=0.3, Ode0=0.70, wp=-0.9, wa=0.2, zp=0.5)
    flatcosmo = FlatwpwaCDM(H0=70, Om0=0.3, wp=-0.9, wa=0.2, zp=0.5)
    assert u.allclose(
        cosmo.de_density_scale(z), flatcosmo.de_density_scale(z), rtol=1e-7
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import io
import os
import os.path
import shutil
import sys
from collections import defaultdict
from os.path import join
from pathlib import Path

from numpy import get_include as get_numpy_include
from setuptools import Extension

from extension_helpers import get_compiler, import_file, pkg_config, write_if_different

WCSROOT = os.path.relpath(os.path.dirname(__file__))
WCSVERSION = "8.3"


def b(s):
    return s.encode("ascii")


def string_escape(s):
    s = s.decode("ascii").encode("ascii", "backslashreplace")
    s = s.replace(b"\n", b"\\n")
    s = s.replace(b"\0", b"\\0")
    return s.decode("ascii")


def determine_64_bit_int():
    """
    The only configuration parameter needed at compile-time is how to
    specify a 64-bit signed integer.  Python's ctypes module can get us
    that information.
    If we can't be absolutely certain, we default to "long long int",
    which is correct on most platforms (x86, x86_64).  If we find
    platforms where this heuristic doesn't work, we may need to
    hardcode for them.
    """
    try:
        try:
            import ctypes
        except ImportError:
            raise ValueError()

        if ctypes.sizeof(ctypes.c_longlong) == 8:
            return "long long int"
        elif ctypes.sizeof(ctypes.c_long) == 8:
            return "long int"
        elif ctypes.sizeof(ctypes.c_int) == 8:
            return "int"
        else:
            raise ValueError()

    except ValueError:
        return "long long int"


def write_wcsconfig_h(paths):
    """
    Writes out the wcsconfig.h header with local configuration.
    """
    h_file = io.StringIO()
    h_file.write(
        f"""
    /* The bundled version has WCSLIB_VERSION */
    #define HAVE_WCSLIB_VERSION 1

    /* WCSLIB library version number. */
    #define WCSLIB_VERSION {WCSVERSION}

    /* 64-bit integer data type. */
    #define WCSLIB_INT64 {determine_64_bit_int()}

    /* Windows needs some other defines to prevent inclusion of wcsset()
       which conflicts with wcslib's wcsset().  These need to be set
       on code that *uses* astropy.wcs, in addition to astropy.wcs itself.
       */
    #if defined(_WIN32) || defined(_MSC_VER) || defined(__MINGW32__) || defined (__MINGW64__)

    #ifndef YY_NO_UNISTD_H
    #define YY_NO_UNISTD_H
    #endif

    #ifndef _CRT_SECURE_NO_WARNINGS
    #define _CRT_SECURE_NO_WARNINGS
    #endif

    #ifndef _NO_OLDNAMES
    #define _NO_OLDNAMES
    #endif

    #ifndef NO_OLDNAMES
    #define NO_OLDNAMES
    #endif

    #ifndef __STDC__
    #define __STDC__ 1
    #endif

    #endif
    """
    )
    content = h_file.getvalue().encode("ascii")
    for path in paths:
        write_if_different(path, content)


######################################################################
# GENERATE DOCSTRINGS IN C


def generate_c_docstrings():
    docstrings = import_file(os.path.join(WCSROOT, "docstrings.py"))
    docstrings = docstrings.__dict__
    keys = [
        key for key, val in docstrings.items()
        if not key.startswith("__") and isinstance(val, str)
    ]  # fmt: skip
    keys.sort()
    docs = {}
    for key in keys:
        docs[key] = docstrings[key].encode("utf8").lstrip() + b"\0"

    h_file = io.StringIO()
    h_file.write(
        """/*
DO NOT EDIT!

This file is autogenerated by astropy/wcs/setup_package.py.  To edit
its contents, edit astropy/wcs/docstrings.py
*/

#ifndef __DOCSTRINGS_H__
#define __DOCSTRINGS_H__

"""
    )
    for key in keys:
        val = docs[key]
        h_file.write(f"extern char doc_{key}[{len(val)}];\n")
    h_file.write("\n#endif\n\n")

    write_if_different(
        join(WCSROOT, "include", "astropy_wcs", "docstrings.h"),
        h_file.getvalue().encode("utf-8"),
    )

    c_file = io.StringIO()
    c_file.write(
        """/*
DO NOT EDIT!

This file is autogenerated by astropy/wcs/setup_package.py.  To edit
its contents, edit astropy/wcs/docstrings.py

The weirdness here with strncpy is because some C compilers, notably
MSVC, do not support string literals greater than 256 characters.
*/

#include <string.h>
#include "astropy_wcs/docstrings.h"

"""
    )
    for key in keys:
        val = docs[key]
        c_file.write(f"char doc_{key}[{len(val)}] = {{\n")
        for i in range(0, len(val), 12):
            section = val[i : i + 12]
            c_file.write("    ")
            c_file.write("".join(f"0x{x:02x}, " for x in section))
            c_file.write("\n")

        c_file.write("    };\n\n")

    write_if_different(
        join(WCSROOT, "src", "docstrings.c"), c_file.getvalue().encode("utf-8")
    )


def get_wcslib_cfg(cfg, wcslib_files, include_paths):
    debug = "--debug" in sys.argv

    cfg["include_dirs"].append(get_numpy_include())
    cfg["define_macros"].extend(
        [
            ("ECHO", None),
            ("WCSTRIG_MACRO", None),
            ("ASTROPY_WCS_BUILD", None),
            ("_GNU_SOURCE", None),
        ]
    )

    if (
        int(os.environ.get("ASTROPY_USE_SYSTEM_WCSLIB", 0))
        or int(os.environ.get("ASTROPY_USE_SYSTEM_ALL", 0))
    ) and not sys.platform == "win32":
        wcsconfig_h_path = join(WCSROOT, "include", "wcsconfig.h")
        if os.path.exists(wcsconfig_h_path):
            os.unlink(wcsconfig_h_path)
        for k, v in pkg_config(["wcslib"], ["wcs"]).items():
            cfg[k].extend(v)
    else:
        write_wcsconfig_h(include_paths)

        wcslib_path = join("cextern", "wcslib")  # Path to wcslib
        wcslib_cpath = join(wcslib_path, "C")  # Path to wcslib source files
        cfg["sources"].extend(join(wcslib_cpath, x) for x in wcslib_files)
        cfg["include_dirs"].append(wcslib_cpath)

    if debug:
        cfg["define_macros"].append(("DEBUG", None))
        cfg["undef_macros"].append("NDEBUG")
        if not sys.platform.startswith("sun") and not sys.platform == "win32":
            cfg["extra_compile_args"].extend(["-fno-inline", "-O0", "-g"])
    else:
        # Define ECHO as nothing to prevent spurious newlines from
        # printing within the libwcs parser
        cfg["define_macros"].append(("NDEBUG", None))
        cfg["undef_macros"].append("DEBUG")

    if sys.platform == "win32":
        # These are written into wcsconfig.h, but that file is not
        # used by all parts of wcslib.
        cfg["define_macros"].extend(
            [
                ("YY_NO_UNISTD_H", None),
                ("_CRT_SECURE_NO_WARNINGS", None),
                ("_NO_OLDNAMES", None),  # for mingw32
                ("NO_OLDNAMES", None),  # for mingw64
                ("__STDC__", None),  # for MSVC
            ]
        )

    if sys.platform.startswith("linux"):
        cfg["define_macros"].append(("HAVE_SINCOS", None))

    # For 4.7+ enable C99 syntax in older compilers (need 'gnu99' std for gcc)
    if get_compiler() == "unix":
        cfg["extra_compile_args"].extend(["-std=gnu99"])

    # Squelch a few compilation warnings in WCSLIB
    if get_compiler() in ("unix", "mingw32"):
        if not debug:
            cfg["extra_compile_args"].extend(
                [
                    "-Wno-strict-prototypes",
                    "-Wno-unused-function",
                    "-Wno-unused-value",
                    "-Wno-uninitialized",
                ]
            )


def get_extensions():
    generate_c_docstrings()

    ######################################################################
    # DISTUTILS SETUP
    cfg = defaultdict(list)

    wcslib_files = [  # List of wcslib files to compile
        "flexed/wcsbth.c",
        "flexed/wcspih.c",
        "flexed/wcsulex.c",
        "flexed/wcsutrn.c",
        "cel.c",
        "dis.c",
        "lin.c",
        "log.c",
        "prj.c",
        "spc.c",
        "sph.c",
        "spx.c",
        "tab.c",
        "wcs.c",
        "wcserr.c",
        "wcsfix.c",
        "wcshdr.c",
        "wcsprintf.c",
        "wcsunits.c",
        "wcsutil.c",
    ]

    wcslib_config_paths = [
        join(WCSROOT, "include", "astropy_wcs", "wcsconfig.h"),
        join(WCSROOT, "include", "wcsconfig.h"),
    ]

    get_wcslib_cfg(cfg, wcslib_files, wcslib_config_paths)

    cfg["include_dirs"].append(join(WCSROOT, "include"))

    astropy_wcs_files = [  # List of astropy.wcs files to compile
        "distortion.c",
        "distortion_wrap.c",
        "docstrings.c",
        "pipeline.c",
        "pyutil.c",
        "astropy_wcs.c",
        "astropy_wcs_api.c",
        "sip.c",
        "sip_wrap.c",
        "str_list_proxy.c",
        "unit_list_proxy.c",
        "util.c",
        "wcslib_wrap.c",
        "wcslib_auxprm_wrap.c",
        "wcslib_prjprm_wrap.c",
        "wcslib_celprm_wrap.c",
        "wcslib_tabprm_wrap.c",
        "wcslib_wtbarr_wrap.c",
    ]
    cfg["sources"].extend(join(WCSROOT, "src", x) for x in astropy_wcs_files)

    cfg["sources"] = [str(x) for x in cfg["sources"]]
    cfg = {str(key): val for key, val in cfg.items()}

    # Copy over header files from WCSLIB into the installed version of Astropy
    # so that other Python packages can write extensions that link to it. We
    # do the copying here then include the data in [tools.setuptools.package_data]
    # in the pyproject.toml file

    wcslib_headers = [
        "cel.h",
        "lin.h",
        "prj.h",
        "spc.h",
        "spx.h",
        "tab.h",
        "wcs.h",
        "wcserr.h",
        "wcsmath.h",
        "wcsprintf.h",
    ]

    if not (
        int(os.environ.get("ASTROPY_USE_SYSTEM_WCSLIB", 0))
        or int(os.environ.get("ASTROPY_USE_SYSTEM_ALL", 0))
    ):
        for header in wcslib_headers:
            source = Path("cextern", "wcslib", "C", header)
            dest = Path("astropy", "wcs", "include", "wcslib", header)
            if not dest.is_file() or source.stat().st_mtime > dest.stat().st_mtime:
                shutil.copy(source, dest)

    return [Extension("astropy.wcs._wcs", **cfg)]
